#include <assert.h>
#include <errno.h>
#include <google/protobuf/repeated_field.h>
#include <stdlib.h>
#include <unistd.h>

#include "client.h"
#include "log.h"
#include "server.h"
#include "serverpool.h"
#include "state.pb.h"
#include "util.h"

using namespace std;

ServerPool::ServerPool(int size)
	: servers(new Server[size]),
	  num_servers(size),
	  clients_added(0)
{
}

ServerPool::~ServerPool()
{
	delete[] servers;
}
	
CubemapStateProto ServerPool::serialize()
{
	CubemapStateProto state;

	for (int i = 0; i < num_servers; ++i) {
                CubemapStateProto local_state = servers[i].serialize();

		// The stream state should be identical between the servers, so we only store it once,
		// save for the fds, which we keep around to distribute to the servers after re-exec.
		if (i == 0) {
			state.mutable_streams()->MergeFrom(local_state.streams());
		} else {
			assert(state.streams_size() == local_state.streams_size());
			for (int j = 0; j < local_state.streams_size(); ++j) {
				assert(local_state.streams(j).data_fds_size() == 1);
				state.mutable_streams(j)->add_data_fds(local_state.streams(j).data_fds(0));
			}
		}
		for (int j = 0; j < local_state.clients_size(); ++j) {
			state.add_clients()->MergeFrom(local_state.clients(j));
		}
        }

	return state;
}

void ServerPool::add_client(int sock)
{
	servers[clients_added++ % num_servers].add_client_deferred(sock);
}

void ServerPool::add_client_from_serialized(const ClientProto &client)
{
	servers[clients_added++ % num_servers].add_client_from_serialized(client);
}

void ServerPool::add_stream(const string &stream_id, size_t backlog_size, Stream::Encoding encoding)
{
	for (int i = 0; i < num_servers; ++i) {
		servers[i].add_stream(stream_id, backlog_size, encoding);
	}
}

void ServerPool::add_stream_from_serialized(const StreamProto &stream, const vector<int> &data_fds)
{
	assert(!data_fds.empty());
	string contents;
	for (int i = 0; i < num_servers; ++i) {
		int data_fd;
		if (i < int(data_fds.size())) {
			// Reuse one of the existing file descriptors.
			data_fd = data_fds[i];
		} else {
			// Clone the first one.
			if (contents.empty()) {
				if (!read_tempfile(data_fds[0], &contents)) {
					exit(1);
				}
			}
			data_fd = make_tempfile(contents);
		}

		servers[i].add_stream_from_serialized(stream, data_fd);
	}

	// Close and delete any leftovers, if the number of servers was reduced.
	for (size_t i = num_servers; i < data_fds.size(); ++i) {
		int ret;
		do {
			ret = close(data_fds[i]);  // Implicitly deletes the file.
		} while (ret == -1 && errno == EINTR);

		if (ret == -1) {
			log_perror("close");
			// Can still continue.
		}
	}
}

void ServerPool::set_header(const string &stream_id, const string &http_header, const string &stream_header)
{
	for (int i = 0; i < num_servers; ++i) {
		servers[i].set_header(stream_id, http_header, stream_header);
	}
}

void ServerPool::add_data(const string &stream_id, const char *data, size_t bytes)
{
	for (int i = 0; i < num_servers; ++i) {
		servers[i].add_data_deferred(stream_id, data, bytes);
	}
}

void ServerPool::run()
{
	for (int i = 0; i < num_servers; ++i) {
		servers[i].run();
	}
}
	
void ServerPool::stop()
{
	for (int i = 0; i < num_servers; ++i) {
		servers[i].stop();
	}
}
	
vector<ClientStats> ServerPool::get_client_stats() const
{
	vector<ClientStats> ret;
	for (int i = 0; i < num_servers; ++i) {
		vector<ClientStats> stats = servers[i].get_client_stats();
		ret.insert(ret.end(), stats.begin(), stats.end());
	}
	return ret;
}
	
void ServerPool::set_mark_pool(const string &stream_id, MarkPool *mark_pool)
{
	for (int i = 0; i < num_servers; ++i) {
		servers[i].set_mark_pool(stream_id, mark_pool);
	}	
}

void ServerPool::set_backlog_size(const string &stream_id, size_t new_size)
{
	for (int i = 0; i < num_servers; ++i) {
		servers[i].set_backlog_size(stream_id, new_size);
	}	
}

void ServerPool::set_encoding(const string &stream_id, Stream::Encoding encoding)
{
	for (int i = 0; i < num_servers; ++i) {
		servers[i].set_encoding(stream_id, encoding);
	}	
}
